package com.sillycycle.bagleyd.tetris;

/*
 * @(#)ThingView.java
 *
 * Copyright 1993 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;

import com.sillycycle.bagleyd.util.Colour;

import android.graphics.Canvas;
import android.graphics.Point;

/**
 * The <code>Thing</code> class keeps track of the falling shapes.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class ThingView {
	//int             units, polyNumber;
	//Point           position = new Point(0, 0);
	//int             size, colorNumber, direction;
	//int             randomRotation, randomReflection;
	//int             randomNumber;
	int             background;
	//boolean         bonus;

	//private GameView gameView;
	//private Bitmap[] bmps;
	int boxSize;
	int boxWidth = boxSize, boxHeight = boxSize;
	Point fieldOffset = new Point(2, 2);
	int columns = TetrisInterface.DEFAULT_COLUMNS, rows = TetrisInterface.DEFAULT_ROWS;
	private Thing curThing;
	private Spec curSpec = new Spec(TetrisInterface.DEFAULT_UNITS);
	//private PolyominoImage polyominoImage = new PolyominoImage();
	private PolyMode[] tris; //[TetrisInterface.MAX_UNITS - MIN_UNITS + 1];
	private Field[][] field = new Field[rows][columns];
	Random generator;
	private boolean suddenAppear = false;

	public ThingView(/*GameView gameView, Bitmap[] bmps*/) {
		/*this.gameView = gameView;
		this.bmps = bmps;*/
		curThing = new Thing();
		generator = new Random(System.nanoTime());
		setCurThing();
		for (int j = 0; j < rows; j++) {
			for (int i = 0; i < columns; i++) {
				field[j][i] = new Field();
				field[j][i].pmid = -1;
				field[j][i].cid = 0;
			}
		}
	}

	public void setBoxSize(int size) {
		boxWidth = boxHeight = boxSize = size;
	}

	public int getBoxSize() {
		return boxSize;
	}

	public int getFirstX() {
		if (tris == null)
			System.out.println("tris = null");
		if (tris[curThing.units].poly == null)
			System.out.println("tris.poly = null");
		if (tris[curThing.units].poly[curThing.polyNumber].shape == null)
			System.out.println("tris.poly.shape = null");
		for (int i = 0; i < curThing.size; i++)
			for (int j = 0; j < curThing.size; j++) {

				 if (tris[curThing.units].poly[curThing.polyNumber].shape[j][i] != 0) {
					 return i;
				 }
			}
		return 0;
	}

	public int getLastX() {
		for (int i = curThing.size - 1; i >= 0; i--)
			for (int j = 0; j < curThing.size; j++) {
				 if (tris[curThing.units].poly[curThing.polyNumber].shape[j][i] != 0) {
					 return i;
				 }
			}
		return 0;
	}
	public int getFirstY() {
		for (int j = 0; j < curThing.size; j++)
			for (int i = 0; i < curThing.size; i++) {
				 if (tris[curThing.units].poly[curThing.polyNumber].shape[j][i] != 0) {
					 return j;
				 }
			}
		return 0;
	}

	public int getLastY() {
		for (int j = curThing.size - 1; j >= 0; j--)
			for (int i = 0; i < curThing.size; i++) {
				 if (tris[curThing.units].poly[curThing.polyNumber].shape[j][i] != 0) {
					 return j;
				 }
			}
		return 0;
	}

	public void setOffset(int offsetX, int offsetY) {
		fieldOffset.x = offsetX;
		fieldOffset.y = offsetY;
	}

	public int getOffsetX() {
		return fieldOffset.x;
	}

	public int getOffsetY() {
		return fieldOffset.y;
	}

	public void setPosition(int x, int y) {
		curThing.position.x = x;
		curThing.position.y = y;
	}

	public void setX(int x) {
		curThing.position.x = x;
	}

	public void setY(int y) {
		curThing.position.y = y;
	}

	public int getX() {
		return curThing.position.x;
	}

	public int getY() {
		return curThing.position.y;
	}

	public final void setCurThing() {
		tris = new PolyMode[TetrisInterface.MAX_UNITS -
                        TetrisInterface.MIN_UNITS + 1];
		for (int i = 0; i < (TetrisInterface.MAX_UNITS -
                                TetrisInterface.MIN_UNITS + 1); i++)
			tris[i] = new PolyMode(TetrisInterface.MAX_MODES,
                                TetrisInterface.MAX_TYPES,
                                TetrisInterface.MAX_START_POLYOMINOES,
                                TetrisInterface.MAX_POLYOMINOES,
                                TetrisInterface.MAX_UNITS, TetrisInterface.MAX_UNITS);
		readPolyominoes();
		setNextThing();
	}

	public void setNextThing() {
		curThing.setRandomNumber(Math.abs(generator.nextInt()));
		curThing.setRotation(generator.nextInt(TetrisInterface.MAX_SIDES));
		curThing.setReflection(generator.nextInt(2));
		curThing.setBonus(curSpec.bonus);

		int diagonal = (curSpec.diagonal) ? 1 : 0;
		int turnStyle =
		tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
                    [diagonal].turnStyle;
		int curStart;

		curThing.units = curSpec.units - TetrisInterface.MIN_UNITS +
				((curThing.getBonus()) ? 1 : 0);
		curStart = curThing.getRandomNumber() %
				tris[curThing.units].mode[diagonal].number
				[(curThing.getBonus()) ? TetrisInterface.NOMIX :
				0][turnStyle];
		curThing.colorNumber = curStart;

		curThing.polyNumber = tris[curThing.units].mode
				[diagonal].start[curStart][turnStyle];
		curThing.size = tris[curThing.units].poly
				[curThing.polyNumber].size;
		curThing.position.x = TetrisInterface.DEFAULT_COLUMNS / 2 - 1;
		curThing.position.y = -tris[curThing.units].poly
				[curThing.polyNumber].startHeight[((suddenAppear) ? 1 : 0)];
	}

	public void surface() {
		while (!atTop()) {
			curThing.position.y--;
		}
	}

	public void fall() {
		while (!atBottom()) {
			curThing.position.y++;
		}
	}

	public boolean move(int x, int y) {
		curThing.position.x += x;
		curThing.position.y += y;
		if (overlapping()) {
			curThing.position.x -= x;
			curThing.position.y -= y;
			return false;
		}
		return true;
	}

	public boolean rotate() {
		Thing oldThing = new Thing();
		oldThing.clone(curThing);
		rotateThing(true); // cw
		if (overlapping()) {
			curThing.clone(oldThing);
			return false;
		}
		return true;
	}

	public boolean reflect() {
		Thing oldThing = new Thing();
		oldThing.clone(curThing);
		reflectThing();
		if (overlapping()) {
			curThing.clone(oldThing);
			return false;
		}
		return true;
	}

	private void rotateThing(boolean cw) {
		if (tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
				[(curSpec.diagonal) ? 1 : 0].turnStyle > TetrisInterface.NONE) {
			  if (cw)
			    for (int i = 0; i < TetrisInterface.MAX_SIDES - 1; i++)
			      curThing.polyNumber = tris[curThing.units].poly
				[curThing.polyNumber].rotation;
			  else /* ccw */
			    curThing.polyNumber = tris[curThing.units].poly
			      [curThing.polyNumber].rotation;
		}
	}

	private void reflectThing() {
		/* reflect on y axis */
		if (tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
		    [(curSpec.diagonal) ? 1 : 0].turnStyle == TetrisInterface.ALL) {
		  curThing.polyNumber = tris[curThing.units].poly
		      [curThing.polyNumber].reflection;
		}
	}

    private void update() {
	// move thing
        /*if (x >= gameView.getWidth() - width - xSpeed || x + xSpeed <= 0) {
               xSpeed = -xSpeed;
        }
        x = x + xSpeed;
        if (y >= gameView.getHeight() - height - ySpeed || y + ySpeed <= 0) {
               ySpeed = -ySpeed;
        }
        y = y + ySpeed;
        currentFrame = ++currentFrame % BMP_COLUMNS;*/
    }

    public void onDraw(Canvas canvas) {
        update();
        /*int srcX = currentFrame * width;
        int srcY = getAnimationRow() * height;
        Rect src = new Rect(srcX, srcY, srcX + width, srcY + height);
        Rect dst = new Rect(x, y, x + width, y + height);
        canvas.drawBitmap(bmp, src, dst, null);*/
    }

    /*private int getAnimationRow() {
        double dirDouble = (Math.atan2(xSpeed, ySpeed) / (Math.PI / 2) + 2);
        int direction = (int) Math.round(dirDouble) % BMP_ROWS;
        return DIRECTION_TO_ANIMATION_MAP[direction];
    }*/

    /*public boolean isCollision(float x2, float y2) {
        return x2 > x && x2 < x + width
		&& y2 > y && y2 < y + height;
    }*/

	void drawBorder(Canvas canvas) {
		PolyominoImage.drawBorder(canvas, fieldOffset.x, fieldOffset.y,
			fieldOffset.x + TetrisInterface.DEFAULT_COLUMNS * boxWidth,
			fieldOffset.y + TetrisInterface.DEFAULT_ROWS * boxHeight);
	}

	void clearUnit(Canvas canvas, int i, int j) {
		PolyominoImage.clearUnits(canvas, new Colour(background),
			fieldOffset.x, fieldOffset.y,
			i, j, boxWidth, boxHeight);
	}

	void xorUnit(Canvas canvas, PolyominoImage polyominoImage,
			int i, int j) {
		if (field[j][i].pmid >= 0) {
			int color = PolyominoImage.colorThings[field[j][i].cid].getColor();
			{ // Simulate Xor
				int mask = 0xFFFFFF & color;
				color = color - mask + 0xFFFFFFFF - mask;
			}
			polyominoImage.drawUnit(canvas, false, field[j][i].pmid,
				new Colour(color), fieldOffset.x, fieldOffset.y,
				i, j, boxWidth, boxHeight);
		}
	}

	void drawField(Canvas canvas, PolyominoImage polyominoImage,
			int start, int finish) {
		int i, j;

		for (i = 0; i < columns; i++)
			for (j = start; j < finish; j++) {
				if (field[j][i].pmid >= 0)
					polyominoImage.drawUnit(canvas, false,
						field[j][i].pmid,
						field[j][i].cid,
						fieldOffset.x, fieldOffset.y,
						i, j, boxWidth, boxHeight);
				else
					clearUnit(canvas, i, j);
			}
	}

	void drawField(Canvas canvas, PolyominoImage polyominoImage) {
		drawField(canvas, polyominoImage, 0, rows);
	}

	void drawThing(Canvas canvas, PolyominoImage polyominoImage) {
		int i, j;

		int[][] shape = tris[curThing.units].poly[curThing.polyNumber].shape;
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = curThing.position.x + i;
				int yj = curThing.position.y + j;

				if (yj >= 0 && tris[curThing.units].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  polyominoImage.drawUnit(canvas, false,
				    PolyominoImage.polyPixmap(i, j,
				      shape,
				      curThing.size),
				      curThing.colorNumber,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight);
				  /* This uses precomputed values */
				  /* polyominoImage.drawUnit(canvas, false,
				    PolyominoImage.NUM_BITMAPS - tris[curThing.units].poly
				      [curThing.polyNumber].shape[j][i] - 1,
				      curThing.colorNumb:er,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight); */
				}
			}
	}

	void drawThingDiff(Canvas canvas, Thing oldThing, PolyominoImage polyominoImage) {
		int i, j;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = curThing.position.x + i;
				int yj = curThing.position.y + j;

				if (yj >= 0 && (tris[curThing.units].poly
				       [curThing.polyNumber].shape[j][i] != 0))
				  polyominoImage.drawUnit(canvas, false,
						PolyominoImage.polyPixmap(i, j,
				    tris[curThing.units].poly[curThing.polyNumber].shape,
				    curThing.size),
				    curThing.colorNumber,
				    fieldOffset.x, fieldOffset.y,
				    xi, yj, boxWidth, boxHeight);
				/* This uses precomputed values */
				/*  polyominoImage.drawUnit(g, false,
				    PolyominoImage.NUM_BITMAPS - tris[curThing.units].poly
				      [curThing.polyNumber].shape[j][i] - 1,
				      curThing.colorNumber,
				      fieldOffset.x, fieldOffset.y,
				      xi, yj, boxWidth, boxHeight); */
			}
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				int xi = oldThing.position.x + i;
				int yj = oldThing.position.y + j;
				int ox = xi - curThing.position.x;
				int oy = yj - curThing.position.y;

				if (yj >= 0 && (tris[oldThing.units].poly
				     [oldThing.polyNumber].shape[j][i] != 0) &&
				    ((ox < 0) || (ox >= curThing.size) ||
				     (oy < 0) || (oy >= curThing.size) ||
				     tris[curThing.units].poly
				      [curThing.polyNumber].shape[oy][ox] == 0))
					clearUnit(canvas, xi, yj);
			}
	}

	void putUnit() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
				xi = x + i;
				yj = y + j;
				if ((yj >= 0) &&
				    (tris[curThing.units].poly
				     [curThing.polyNumber].shape[j][i] != 0)) {
					field[yj][xi].pmid =
					  PolyominoImage.polyPixmap(i, j, tris[curThing.units].poly
					    [curThing.polyNumber].shape,
					    curThing.size) % PolyominoImage.NUM_BITMAPS;
					/* This uses precomputed values */
					/* field[yj][xi].pmid =
					  PolyominoImage.NUM_BITMAPS - tris[curThing.units].poly
					   [curThing.polyNumber].shape[j][i] - 1;
					*/
					field[yj][xi].cid = curThing.colorNumber;
				}
			}
	}

	boolean overlapping() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
			  if (tris[curThing.units].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    xi = x + i;
			    yj = y + j;
			    if ((yj >= rows) || (xi < 0) || (xi >= columns)) {
			      return true;
			    }
			    if (suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
			      if ((yj < 0) || (field != null && field[yj][xi].pmid >= 0)) {
				  return true;
			      }
			    } else { /* evolution */
/* This method one can turn polyomino to an area above of screen, also
   part of the polyomino may not be visible initially */
			      if ((yj >= 0) && (field != null && field[yj][xi].pmid >= 0)) {
				  return true;
			      }
			    }
			  }
		return false;
	}

	boolean atTop() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
			  xi = x + i;
			  yj = y + j;
			  if ((yj >= -1) && tris[curThing.units].poly
			      [curThing.polyNumber].shape[j][i] != 0 &&
			      ((yj <= 0) ||
			       (xi < 0) || (xi >= columns) ||
			       (field != null && field[yj - 1][xi].pmid >= 0)))
			    return true;
			}
		return false;
	}

	boolean atBottom() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++) {
			  xi = x + i;
			  yj = y + j;
			  if ((yj >= -1) && tris[curThing.units].poly
			      [curThing.polyNumber].shape[j][i] != 0 &&
			      ((yj >= rows - 1) ||
			       (xi < 0) || (xi >= columns) ||
			       (field != null && field[yj + 1][xi].pmid >= 0)))
			    return true;
			}
		return false;
	}

	void initFill(boolean fixedFill, int fill) {
		int i, j, col, fixedUnits;
		int turnStyle = tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int numberPolyominoes = tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0][turnStyle];

		if (fixedFill) {
			/* This algorithm randomly picks fixedUnits per row
			 * (4 for DEFAULTWIDTH, ie. 10/2-1)
			 * This is more in keeping with the original */
			fixedUnits = columns / 2 - 1;
			for (j = 0; j < rows; j++) {
				for (i = 0; i < columns; i++) {
					field[j][i].pmid = -1;
					field[j][i].cid = 0;
				}
				if (j >= rows - fill)
					for (i = 0; i < fixedUnits; i++) {
						do
							col = generator.nextInt(columns);
						while (field[j][col].pmid != -1);
						field[j][col].pmid = PolyominoImage.NUM_BITMAPS - 1;
						field[j][col].cid = generator.nextInt(numberPolyominoes);

					}
			}
		} else {
			/* This algorithm randomly fills boxes to a certain level
			   The problem with this is that it could potentially fill
			   up a whole row */
			for (i = 0; i < columns; i++)
				for (j = 0; j < rows; j++)
					if ((j >= rows - fill) &&
							(generator.nextInt(2) == 0)) {
						field[j][i].pmid = PolyominoImage.NUM_BITMAPS - 1;
						field[j][i].cid = generator.nextInt(numberPolyominoes);
					} else {
						field[j][i].pmid = -1;
						field[j][i].cid = 0;
					}
		}
	}

	void fillLines(Canvas canvas, PolyominoImage polyominoImage, int j) {
		int i;
		int turnStyle = tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int pmid = PolyominoImage.NUM_BITMAPS - 1;

			for (i = 0; i < columns; i++) {
				int cid = generator.nextInt(tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
						  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
								  [turnStyle]);
				field[j][i].pmid = pmid;
				field[j][i].cid = cid;
				polyominoImage.drawUnit(canvas, false,
				  pmid, cid, fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
				cid = generator.nextInt(tris[curSpec.units - TetrisInterface.MIN_UNITS].mode
						  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
								  [turnStyle]);
				field[rows - j - 1][i].pmid = pmid;
				field[rows - j - 1][i].cid = cid;
				polyominoImage.drawUnit(canvas, false,
				  pmid, cid, fieldOffset.x, fieldOffset.y,
				  i, rows - j - 1, boxWidth, boxHeight);
			}
	}

	int checkLines() {
		int[] lSet = new int[rows];
		int nSet = 0, i, j;

		/* Find filled rows */
		for (j = 0; j < rows; j++) {
			lSet[j] = 0;
			for (i = 0; i < columns; i++)
				if (field[j][i].pmid >= 0)
					lSet[j]++;
			if (lSet[j] == columns)
				nSet++;
		}
		return nSet;
	}

	int invertLines(Canvas canvas, PolyominoImage polyominoImage) {
		int[] lSet = new int[rows];
		int nSet = 0, i, j;

		/* Find filled rows */
		for (j = 0; j < rows; j++) {
			lSet[j] = 0;
			for (i = 0; i < columns; i++)
				if (field[j][i].pmid >= 0)
					lSet[j]++;
			if (lSet[j] == columns)
				nSet++;
		}
		if (nSet == 0) {
			return 0;
		}
		  for (j = 0; j < rows; j++) {
		    if (lSet[j] == columns)
		      for (i = 0; i < columns; i++) {
			  xorUnit(canvas, polyominoImage, i, j);
		      }
		  }
		return nSet;
	}

	int shrinkLines(Canvas canvas, PolyominoImage polyominoImage) {
		int[] lSet = new int[rows];
		int nSet = 0, i, j, y;

		/* Find filled rows */
		for (j = 0; j < rows; j++) {
			lSet[j] = 0;
			for (i = 0; i < columns; i++)
				if (field[j][i].pmid >= 0)
					lSet[j]++;
			if (lSet[j] == columns)
				nSet++;
		}
		if (nSet == 0) {
			return 0;
		}
		/* Shrink */
		for (j = rows - 1; j >= 0; j--)
			if (lSet[j] == columns) {
			  for (y = j; y > 0; y--)
			    for (i = 0; i < columns; i++) {
			      field[y][i].pmid = field[y - 1][i].pmid;
			      field[y][i].cid = field[y - 1][i].cid;
			    }
			  for (i = 0; i < columns; i++)
			    field[0][i].pmid = -1;
			  drawField(canvas, polyominoImage, 0, j + 1);

			  for (i = j; i > 0; i--)
			    lSet[i] = lSet[i - 1];
			  lSet[0] = 0;

			  if (j > 0)
			    for (i = 0; i < columns; i++) {
			      int tmp = field[j][i].pmid;
			      if ((tmp >= 0) && (tmp != PolyominoImage.CHECKDOWN(tmp))) {
				field[j][i].pmid = PolyominoImage.CHECKDOWN(tmp);
				polyominoImage.drawUnit(canvas, false,
				  field[j][i].pmid, field[j][i].cid,
				  fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
			      }
			    }
			  j++;
			  if (j < rows)
			    for (i = 0; i < columns; i++) {
			      int tmp = field[j][i].pmid;

			      if ((tmp >= 0) && (tmp != PolyominoImage.CHECKUP(tmp))) {
				field[j][i].pmid = PolyominoImage.CHECKUP(tmp);
				polyominoImage.drawUnit(canvas, false,
				  field[j][i].pmid, field[j][i].cid,
				  fieldOffset.x, fieldOffset.y,
				  i, j, boxWidth, boxHeight);
			      }
			    }
			}
			//bumpSound
		return nSet;
	}

	int lc, cc;
	char c;

	char readChar() {
		int lcold = lc, ccold = cc;

		if (lc >= polyominoData.length)
			return ' ';
		if (cc == polyominoData[lc].length()) {
			lc++;
			cc = 0;
		} else
			cc++;
		if (ccold >= polyominoData[lcold].length())
			return ' ';
		return polyominoData[lcold].charAt(ccold);
	}

	int readInt() {
		boolean first = false;
		int loopcount = 0;
		int number;
		int MAX_LOOP = 100;

		while (!first) {
			c = readChar();
			if (c >= '0' && c <= '9')
				first = true;
			if (loopcount++ >= MAX_LOOP) {
				System.err.println(
					"Tetris: readInt1 infinite loop.");

			}
		}
		number = c - '0';
		first = false;
		loopcount = 0;
		while (!first) {
			int BASE = 10;
			c = readChar();
			if (c < '0' || c > '9')
				first = true;
			else
				number = BASE * number + (c - '0');
			if (loopcount++ >= MAX_LOOP) {
				System.err.println(
					"Tetris: readInt2 infinite loop.");
			}
		}
		return number;
	}

	void readPolyominoes() {
		int i, j, k, sq, polyomino, sum, start, n, size = 0;
		int height, diag, initHeight, toss = 0;
		int[] counter = new int [TetrisInterface.MAX_UNITS -
			TetrisInterface.MIN_UNITS + 1];
		int[][][] startCounter = new int [TetrisInterface.MAX_UNITS -
			TetrisInterface.MIN_UNITS + 1]
			[TetrisInterface.MAX_MODES][TetrisInterface.MAX_TYPES];

		lc = 0; cc = 0;
		for (sq = 0; sq <= TetrisInterface.MAX_UNITS -
		     TetrisInterface.MIN_UNITS; sq++) {
		  counter[sq] = 0;
		  for (polyomino = 0; polyomino <
		       TetrisInterface.MAX_POLYOMINOES; polyomino++)
		    for (j = 0; j < TetrisInterface.MAX_UNITS; j++)
		      for (i = 0; i < TetrisInterface.MAX_UNITS; i++)
			tris[sq].poly[polyomino].shape[j][i] = 0;
		  for (j = 0; j < TetrisInterface.MAX_TYPES; j++) {
		    for (polyomino = 0; polyomino <
			 TetrisInterface.MAX_START_POLYOMINOES; polyomino++) {
		      tris[sq].mode[TetrisInterface.NODIAG].start
			 [polyomino][j] = 0;
		      tris[sq].mode[TetrisInterface.DIAGONAL].start
			 [polyomino][j] = 0;
		    }
		    startCounter[sq][TetrisInterface.NODIAG][j] = 0;
		    startCounter[sq][TetrisInterface.DIAGONAL][j] = 0;
		    tris[sq].mode[TetrisInterface.NODIAG].number
		      [TetrisInterface.NOMIX][j] = 0;
		    tris[sq].mode[TetrisInterface.DIAGONAL].number
		      [TetrisInterface.NOMIX][j] = 0;
		    tris[sq].mode[TetrisInterface.NODIAG].number
		      [TetrisInterface.MIXED][j] = 0;
		    tris[sq].mode[TetrisInterface.DIAGONAL].number
		      [TetrisInterface.MIXED][j] = 0;
		  }
		  tris[sq].mode[TetrisInterface.NODIAG].turnStyle =
		    TetrisInterface.NONE;
		  tris[sq].mode[TetrisInterface.DIAGONAL].turnStyle =
		    TetrisInterface.NONE;
		}
		c = readChar();
		while (lc + 1 < polyominoData.length) {
		  if (c == '+') {
		    k = readInt();
		    sq = k - TetrisInterface.MIN_UNITS;
		    for (i = 0; i < TetrisInterface.MAX_TYPES; i++) {
		      k = readInt();
		      tris[sq].mode[TetrisInterface.NODIAG].number
			[TetrisInterface.NOMIX][i] = k;
		    }
		    k = readInt();
		    tris[sq].mode[TetrisInterface.NODIAG].turnStyle = k;
		    k = readInt();
		    tris[sq].diagonalSwitch = k;
		    if (tris[sq].diagonalSwitch != 0) {
		      for (i = 0; i < TetrisInterface.MAX_TYPES; i++) {
			k = readInt();
			tris[sq].mode[TetrisInterface.DIAGONAL].number
			  [TetrisInterface.NOMIX][i] = k;
		      }
		      k = readInt();
		      tris[sq].mode[TetrisInterface.DIAGONAL].turnStyle = k;
		    }
		    c = readChar();
		  }
		  if (c == '*') {
		    size = readInt();
		    c = readChar();
		  }
		  /* Useful for debugging Things */
/*
		  if (c == '~') {
		    toss = readInt(&lc, &cc, &c);
		  }
*/
		  if (c == '#') {
		    k = readInt();
		    sq = k - TetrisInterface.MIN_UNITS;
		    n = readInt();
		    if (tris[sq].diagonalSwitch != 0)
		      n = readInt();
		    for (polyomino = 0; polyomino < n - toss; polyomino++) {
		      sum = polyomino + counter[sq];
		      /* This is only there to "read" input file */
		      readInt();
		      k = readInt();
		      tris[sq].poly[sum].rotation =
			k + counter[sq] - toss;
		      k = readInt();
		      tris[sq].poly[sum].reflection =
			k + counter[sq] - toss;
		      for (initHeight = TetrisInterface.JUMPIN;
			   initHeight <= TetrisInterface.GRADUAL; initHeight++) {
			height = readInt();
			if (initHeight == TetrisInterface.JUMPIN)
			  tris[sq].poly[sum].startHeight[1] = height;
			else if (initHeight == TetrisInterface.GRADUAL)
			  tris[sq].poly[sum].startHeight[0] = height;
			for (diag = TetrisInterface.NODIAG;
			     diag <= tris[sq].diagonalSwitch; diag++) {
			  start = readInt();
			  if (initHeight == TetrisInterface.JUMPIN) {
			    if (sq == 0 ||
				tris[sq - 1].mode[diag].turnStyle ==
				TetrisInterface.NONE) {
			      i = startCounter[sq][diag][TetrisInterface.NONE];
			      tris[sq].mode[diag].start[i]
				 [TetrisInterface.NONE] = sum;
			      startCounter[sq][diag][TetrisInterface.NONE]++;
			    }
			    if ((sq == 0 ||
				tris[sq - 1].mode[diag].turnStyle <
				TetrisInterface.ALL) && start !=
				TetrisInterface.NONE) {
			      i = startCounter[sq][diag]
				[TetrisInterface.NOREFL];
			      tris[sq].mode[diag].start[i]
				[TetrisInterface.NOREFL] = sum;
			      startCounter[sq][diag][TetrisInterface.NOREFL]++;
			    }
			    if (start == TetrisInterface.ALL) {
			      i= startCounter[sq][diag][TetrisInterface.ALL];
			      tris[sq].mode[diag].start[i]
				[TetrisInterface.ALL] = sum;
			      startCounter[sq][diag][TetrisInterface.ALL]++;
			    }
			  }
			}
		      }
		      tris[sq].poly[sum].size = size;
		      for (j = 0; j < size; j++)
			for (i = 0; i < size; i++) {
			   k = readInt();
			   tris[sq].poly[sum].shape[j][i] = k;
			}
		    }
		    counter[sq] += n - toss;
		    toss = 0;
		    c = readChar();
		  }
		}
		for (i = 0; i <= TetrisInterface.MAX_UNITS -
		     TetrisInterface.MIN_UNITS; i++)
		  for (j = 0; j < TetrisInterface.MAX_TYPES; j++)
		    for (k = 0; k <= i; k++) {
		      tris[i].mode[TetrisInterface.NODIAG].number
			    [TetrisInterface.MIXED][j] +=
			  tris[k].mode[TetrisInterface.NODIAG].number
			    [TetrisInterface.NOMIX][j];
		      if (tris[i].diagonalSwitch != 0)
			/*since k & i are not independent*/
			tris[i].mode[TetrisInterface.DIAGONAL].number
			    [TetrisInterface.MIXED][j] +=
			  tris[k].mode[TetrisInterface.DIAGONAL].number
			    [TetrisInterface.NOMIX][j];
		    }
	}

	class Field {
		int pmid, cid; // pixmaps and colors

		public Field() {
			pmid = -1;
			cid = 0;
		}
	}
/* polyominoes for tetris & welltris */

/* Maximum number of squares was chosen to be 6 because if 7, it leaves a */
/*  hole: @@@ */
/*        @ @ */
/*        @@  */
/*  likewise, 3 was chosen for diagonal mode because if 4, it leaves a hole: */
/*   @  */
/*  @ @ */
/*   @  */

/* '+' Prelude part */
/*   1st number = number of squares in a polyomino */
/*   2nd number = number of distinct polyominoes & rotations & reflections */
/*   3rd number = number of distinct polyominoes & reflections */
/*   4th number = number of distinct polyominoes */
/*   5th number = 0 do not allow rotation and reflection */
/*                1 allow rotation but not reflection */
/*                2 allow rotation and reflection */
/*   6th number = diagonal mode to consider? 0 or 1 (if 0, no 7-10) */
/*   7th number = diagonal number of distinct polyominoes & rots & refls */
/*   8th number = diagonal number of distinct polyominoes & reflections */
/*   9th number = diagonal number of distinct polyominoes */
/*  10th number = 0 do not allow rotation and reflection for diagonal mode */
/*                1 allow rotation but not reflection for diagonal mode */
/*                2 allow rotation and reflection for diagonal mode */

/* '*' Main parts */
/*   1st number = n-> consider an n^2 square */
/* '#' polyomino group (follows a '*') */
/*   1st number = number of squares in a polyomino */
/*   2nd number = number of polyominoes (incl rot & refl) in a n^2 square */
/*   3rd number = if it is there, number of polyominoes (incl rot & refl) */
/*       in a n^2 square with diagonal mode */
/* The polyomino (follows a '#') */
/*   1st number = numbered polyomino (there for your benefit) */
/*   2nd number = rotation pointer (usually the next polyomino) */
/*   3rd number = reflection pointer */
/*   4th number = number of rows down to start */
/*   5th number = start number (0, 1, or 2) */
/*       0=> starter for no turns permitted (no reflections or rotations) */
/*       1=> starter for allow rotations and no turns permitted */
/*       2=> starter for reflections, rotations, and no turns permitted */
/*   Next number (if diagonal mode) = start number for diagonal mode */
/*   The proceeding 2 or 3 numbers are repeated for gradual appearance */
/*       Pieces either JUMPIN or appear GRADUALly */
/*   The last group of n^2 numbers is the actual polyomino */
/*       0=>     empty square */
/*       1..16=> used square */
/* '~' debug (precedes a '#') */
/*   1st number = number of polyominoes to skip */

String[] polyominoData = {

/* PRELUDE */
/* + Polyomino States(None Rot Refl) Style Diagonal [States(N R R) Style] */
/* + 1 1 1 1 0 1     1 1 1 0 */

"+ 2 2 1 1 0 1     4 2 2 0",
"+ 3 6 2 2 0 1     20 6 5 1",
"+ 4 19 7 5 1 0",
"+ 5 63 18 12 2 0",
"+ 6 216 60 35 2 0",

/* MAIN */
/* * X    (polyomino fills X^2 square) */
/* # Polyomino Number_of_Polyominoes_Normal [Number_of_Polyominoes_Diagonal] */
/* Ref_Num Ptr_Rot Ptr_Refl (Start_Location Start_N [Start_D])J&G Shape_Array */
/* * 1 */
/* # 1 1 */
/* 0 0 0  0 2 2 1 2 2   16 */
"* 2",
"# 2 2 4",
"0 1 0  1 2 2 1 0 0   0 0  4 1",
"1 0 1  0 0 0 1 2 2   0 2  0 8",

"2 3 3  0 0 2 1 0 2   0 16  16 0",
"3 2 2  0 0 0 1 0 0   16 0  0 16",
"# 3 4 4",
"0 1 3  0 2 2 1 2 2   6 1  8 0",
"1 2 2  0 0 0 1 0 0   2 0  12 1",
"2 3 1  0 0 0 1 0 0   0 2  4 9",
"3 0 0  0 0 0 1 0 0   4 3  0 8",
"# 4 1",
"0 0 0  0 2 1 2   6 3  12 9",
"* 3",
"# 3 2 16",
"0 1 0   1 2 2 1 0 0   0 0 0  4 5 1   0 0 0",
"1 0 1   0 0 0 2 2 2   0 2 0  0 10 0  0 8 0",

"2 3 3     0 0 2 2 0 2   16 0 0  0 16 0   0 0 16",
"3 2 2     0 0 0 2 0 0   0 0 16  0 16 0   16 0 0",
"4 5 8     1 0 2 2 0 0   0 0 0   0 4 1    16 0 0",
"5 6 11    0 0 0 2 0 2   0 2 0   0 8 0    0 0 16",
"6 7 10    1 0 0 2 0 0   0 0 0   0 0 16   4 1 0",
"7 4 9     0 0 0 2 0 0   0 16 0  0 0 2    0 0 8",
"8 9 4     1 0 1 2 0 0   0 0 0   4 1 0    0 0 16",
"9 10 7    0 0 0 2 0 0   0 0 16  0 2 0    0 8 0",
"10 11 6   1 0 0 2 0 0   0 0 0   16 0 0   0 4 1",
"11 8 5    0 0 0 2 0 1   0 0 2   0 0 8    0 16 0",
"12 13 12  1 0 2 2 0 0   0 0 0   16 0 16  0 16 0",
"13 14 15  0 0 0 2 0 2   0 16 0  0 0 16   0 16 0",
"14 15 14  1 0 0 2 0 0   0 0 0   0 16 0   16 0 16",
"15 12 13  0 0 0 2 0 0   0 0 16  0 16 0   0 0 16",
"# 4 16",
"0 1 4     1 2 2 0   0 0 0  6 5 1   8 0 0",
"1 2 7     0 0 2 0   0 2 0  0 10 0  0 12 1",
"2 3 6     1 0 2 0   0 0 0  0 0 2   4 5 9",
"3 0 5     0 0 2 2   0 4 3  0 0 10  0 0 8",
"4 5 0     1 1 2 0   0 0 0  4 5 3   0 0 8",
"5 6 3     0 0 2 1   0 6 1  0 10 0  0 8 0",
"6 7 2     1 0 2 0   0 0 0  2 0 0   12 5 1",
"7 4 1     0 0 2 0   0 0 2  0 0 10  0 4 9",
"8 9 8     1 2 2 0   0 0 0  4 7 1   0 8 0",
"9 10 11   0 0 2 2   0 2 0  0 14 1  0 8 0",
"10 11 10  1 0 2 0   0 0 0  0 2 0   4 13 1",
"11 8 9    0 0 2 0   0 0 2  0 4 11  0 0 8",
"12 13 14  1 2 2 0   0 0 0  4 3 0   0 12 1",
"13 12 15  0 0 2 2   0 0 2  0 6 9   0 8 0",
"14 15 12  1 1 2 0   0 0 0  0 6 1   4 9 0",
"15 14 13  0 0 2 1   0 2 0  0 12 3  0 0 8",
"# 5 37",
"0 1 0     1 2 2 0   0 0 0  6 5 3    8 0 8",
"1 2 3     0 0 2 2   0 6 1  0 10 0   0 12 1",
"2 3 2     1 0 2 0   0 0 0  2 0 2    12 5 9",
"3 0 1     0 0 2 0   0 4 3  0 0 10   0 4 9",
"4 5 8     1 2 2 0   0 0 0  6 7 1    12 9 0",
"5 6 11    0 0 2 0   0 2 0  0 14 3   0 12 9",
"6 7 10    1 0 2 0   0 0 0  0 6 3    4 13 9",
"7 4 9     0 0 2 2   0 6 3  0 12 11  0 0 8",
"8 9 4     1 1 2 0   0 0 0  4 7 3    0 12 9",
"9 10 7    0 0 2 1   0 6 3  0 14 9   0 8 0",
"10 11 6   1 0 2 0   0 0 0  6 3 0    12 13 1",
"11 8 5    0 0 2 0   0 0 2  0 6 11   0 12 9",
"12 13 12  0 2 2 2   4 7 1  0 10 0   0 8 0",
"13 14 15  0 0 2 0   2 0 0  14 5 1   8 0 0",
"14 15 14  0 0 2 0   0 2 0  0 10 0   4 13 1",
"15 12 13  0 0 2 0   0 0 2  4 5 11   0 0 8",
"16 17 19  0 2 2 2   6 5 1  10 0 0   8 0 0",
"17 18 18  0 0 2 0   2 0 0  10 0 0   12 5 1",
"18 19 17  0 0 2 0   0 0 2  0 0 10   4 5 9",
"19 16 16  0 0 2 0   4 5 3  0 0 10   0 0 8",
"20 21 22  0 2 2 0   2 0 0  12 5 3   0 0 8",
"21 20 23  0 0 2 2   0 6 1  0 10 0   4 9 0",
"22 23 20  0 1 2 0   0 0 2  6 5 9    8 0 0",
"23 22 21  0 0 2 1   4 3 0  0 10 0   0 12 1",
"24 25 25  0 2 2 2   4 3 0  0 12 3   0 0 8",
"25 26 24  0 0 2 0   0 6 1  6 9 0    8 0 0",
"26 27 27  0 0 2 0   2 0 0  12 3 0   0 12 1",
"27 24 26  0 0 2 0   0 0 2  0 6 9    4 9 0",
"28 29 32  0 2 2 0   0 2 0  6 13 1   8 0 0",
"29 30 35  0 0 2 0   0 2 0  4 11 0   0 12 1",
"30 31 34  0 0 2 0   0 0 2  4 7 9    0 8 0",
"31 28 33  0 0 2 2   4 3 0  0 14 1   0 8 0",
"32 33 28  0 1 2 0   0 2 0  4 13 3   0 0 8",
"33 34 31  0 0 2 1   0 6 1  4 11 0   0 8 0",
"34 35 30  0 0 2 0   2 0 0  12 7 1   0 8 0",
"35 32 29  0 0 2 0   0 2 0  0 14 1   4 9 0",
"36 36 36  0 2 2 2   0 2 0  4 15 1   0 8 0",
"# 6 46",
"0 1 4     0 2 2 0   6 5 3  10 0 8   8 0 0",
"1 2 7     0 0 2 0   6 1 0  10 0 0   12 5 1",
"2 3 6     0 0 2 0   0 0 2  2 0 10   12 5 9",
"3 0 5     0 0 2 2   4 5 3  0 0 10   0 4 9",
"4 5 0     0 1 2 0   6 5 3  8 0 10   0 0 8",
"5 6 3     0 0 2 1   6 5 1  10 0 0   12 1 0",
"6 7 2     0 0 2 0   2 0 0  10 0 2   12 5 9",
"7 4 1     0 0 2 0   0 4 3  0 0 10   4 5 9",
"8 9 12    0 2 2 2   6 7 1  12 11 0  0 8 0",
"9 10 15   0 0 2 0   2 0 0  14 7 1   12 9 0",
"10 11 14  0 0 2 0   0 2 0  0 14 3   4 13 9",
"11 8 13   0 0 2 0   0 6 3  4 13 11  0 0 8",
"12 13 8   0 1 2 1   4 7 3  0 14 9   0 8 0",
"13 14 11  0 0 2 0   6 3 0  14 13 1  8 0 0",
"14 15 10  0 0 2 0   0 2 0  6 11 0   12 13 1",
"15 12 9   0 0 2 0   0 0 2  4 7 11   0 12 9",
"16 17 20  0 2 2 0   6 3 0  12 13 3  0 0 8",
"17 18 23  0 0 2 0   0 6 1  6 11 0   12 9 0",
"18 19 22  0 0 2 0   2 0 0  12 7 3   0 12 9",
"19 16 21  0 0 2 2   0 6 3  0 14 9   4 9 0",
"20 21 16  0 1 2 0   0 6 3  6 13 9   8 0 0",
"21 22 19  0 0 2 1   6 3 0  12 11 0  0 12 1",
"22 23 18  0 0 2 0   0 0 2  6 7 9    12 9 0",
"23 20 17  0 0 2 0   4 3 0  0 14 3   0 12 9",
"24 25 28  0 2 2 2   4 7 1  0 10 0   4 9 0",
"25 26 31  0 0 2 0   2 0 0  14 5 3   8 0 8",
"26 27 30  0 0 2 0   0 6 1  0 10 0   4 13 1",
"27 24 29  0 0 2 0   2 0 2  12 5 11  0 0 8",
"28 29 24  0 1 2 1   4 7 1  0 10 0   0 12 1",
"29 30 27  0 0 2 0   2 0 2  14 5 9   8 0 0",
"30 31 26  0 0 2 0   4 3 0  0 10 0   4 13 1",
"31 28 25  0 0 2 0   0 0 2  6 5 11   8 0 8",
"32 33 32  0 2 2 0   2 0 2  12 7 9   0 8 0",
"33 34 35  0 0 2 2   4 3 0  0 14 1   4 9 0",
"34 35 34  0 0 2 0   0 2 0  6 13 3   8 0 8",
"35 32 33  0 0 2 0   0 6 1  4 11 0   0 12 1",
"36 37 39  0 2 2 2   6 7 1  14 9 0   8 0 0",
"37 38 38  0 0 2 0   2 0 0  14 3 0   12 13 1",
"38 39 37  0 0 2 0   0 0 2  0 6 11   4 13 9",
"39 36 36  0 0 2 0   4 7 3  0 12 11  0 0 8",
"40 41 43  0 2 2 2   6 3 0  12 15 1  0 8 0",
"41 42 42  0 0 2 0   0 2 0  6 15 1   12 9 0",
"42 43 41  0 0 2 0   0 2 0  4 15 3   0 12 9",
"43 40 40  0 0 2 0   0 6 3  4 15 9   0 8 0",
"44 45 44  1 2 2 0   0 0 0  6 7 3    12 13 9",
"45 44 45  0 0 2 2   0 6 3  0 14 11  0 12 9",
"* 4",
"# 4 2",
"0 1 0  2 2 2 0   0 0 0 0  0 0 0 0   4 5 5 1   0 0 0 0",
"1 0 1  0 0 3 2   0 0 2 0  0 0 10 0  0 0 10 0  0 0 8 0",
"# 5 24",
"0 1 4     1 2 2 0   0 0 0 0  6 5 5 1   8 0 0 0   0 0 0 0",
"1 2 7     0 0 3 0   0 2 0 0  0 10 0 0  0 10 0 0  0 12 1 0",
"2 3 6     1 0 2 0   0 0 0 0  0 0 0 2   4 5 5 9   0 0 0 0",
"3 0 5     0 0 3 2   0 4 3 0  0 0 10 0  0 0 10 0  0 0 8 0",
"4 5 0     1 1 2 0   0 0 0 0  4 5 5 3   0 0 0 8   0 0 0 0",
"5 6 3     0 0 3 1   0 6 1 0  0 10 0 0  0 10 0 0  0 8 0 0",
"6 7 2     1 0 2 0   0 0 0 0  2 0 0 0   12 5 5 1  0 0 0 0",
"7 4 1     0 0 3 0   0 0 2 0  0 0 10 0  0 0 10 0  0 4 9 0",
"8 9 12    1 2 2 0   0 0 0 0  4 7 5 1   0 8 0 0   0 0 0 0",
"9 10 15   0 0 3 0   0 2 0 0  0 10 0 0  0 14 1 0  0 8 0 0",
"10 11 14  1 0 2 0   0 0 0 0  0 0 2 0   4 5 13 1  0 0 0 0",
"11 8 13   0 0 3 2   0 0 2 0  0 4 11 0  0 0 10 0  0 0 8 0",
"12 13 8   1 1 2 0   0 0 0 0  4 5 7 1   0 0 8 0   0 0 0 0",
"13 14 11  0 0 3 1   0 2 0 0  0 14 1 0  0 10 0 0  0 8 0 0",
"14 15 10  1 0 2 0   0 0 0 0  0 2 0 0   4 13 5 1  0 0 0 0",
"15 12 9   0 0 3 0   0 0 2 0  0 0 10 0  0 4 11 0  0 0 8 0",
"16 17 20  1 2 2 0   0 0 0 0  4 5 3 0   0 0 12 1  0 0 0 0",
"17 18 23  0 0 3 2   0 0 2 0  0 6 9 0   0 10 0 0  0 8 0 0",
"18 19 22  1 0 2 0   0 0 0 0  4 3 0 0   0 12 5 1  0 0 0 0",
"19 16 21  0 0 3 0   0 0 2 0  0 0 10 0  0 6 9 0   0 8 0 0",
"20 21 16  1 1 2 0   0 0 0 0  0 6 5 1   4 9 0 0   0 0 0 0",
"21 22 19  0 0 3 0   0 2 0 0  0 10 0 0  0 12 3 0  0 0 8 0",
"22 23 18  1 0 2 0   0 0 0 0  0 0 6 1   4 5 9 0   0 0 0 0",
"23 20 17  0 0 3 1   0 2 0 0  0 12 3 0  0 0 10 0  0 0 8 0",
"# 6 136",
"0 1 4        1 2 3 0   0 0 0 0  6 5 5 1    10 0 0 0   8 0 0 0",
"1 2 7        0 0 3 0   0 2 0 0  0 10 0 0   0 10 0 0   0 12 5 1",
"2 3 6        1 0 3 0   0 0 0 0  0 0 0 2    0 0 0 10   4 5 5 9",
"3 0 5        0 0 3 2   0 4 5 3  0 0 0 10   0 0 0 10   0 0 0 8",
"4 5 0        1 1 3 0   0 0 0 0  4 5 5 3    0 0 0 10   0 0 0 8",
"5 6 3        0 0 3 1   0 6 5 1  0 10 0 0   0 10 0 0   0 8 0 0",
"6 7 2        1 0 3 0   0 0 0 0  2 0 0 0    10 0 0 0   12 5 5 1",
"7 4 1        0 0 3 0   0 0 0 2  0 0 0 10   0 0 0 10   0 4 5 9",
"8 9 12       1 2 2 0   0 0 0 0  6 7 5 1    12 9 0 0   0 0 0 0",
"9 10 15      0 0 3 0   0 2 0 0  0 10 0 0   0 14 3 0   0 12 9 0",
"10 11 14     1 0 2 0   0 0 0 0  0 0 6 3    4 5 13 9   0 0 0 0",
"11 8 13      0 0 3 2   0 6 3 0  0 12 11 0  0 0 10 0   0 0 8 0",
"12 13 8      1 1 2 0   0 0 0 0  4 5 7 3    0 0 12 9   0 0 0 0",
"13 14 11     0 0 3 1   0 6 3 0  0 14 9 0   0 10 0 0   0 8 0 0",
"14 15 10     1 0 2 0   0 0 0 0  6 3 0 0    12 13 5 1  0 0 0 0",
"15 12 9      0 0 3 0   0 0 2 0  0 0 10 0   0 6 11 0   0 12 9 0",
"16 17 20     1 2 2 0   0 0 0 0  6 5 7 1    8 0 8 0    0 0 0 0",
"17 18 23     0 0 3 0   0 2 0 0  0 14 1 0   0 10 0 0   0 12 1 0",
"18 19 22     1 0 2 0   0 0 0 0  0 2 0 2    4 13 5 9   0 0 0 0",
"19 16 21     0 0 3 2   0 4 3 0  0 0 10 0   0 4 11 0   0 0 8 0",
"20 21 16     1 1 2 0   0 0 0 0  4 7 5 3    0 8 0 8    0 0 0 0",
"21 22 19     0 0 3 1   0 6 1 0  0 10 0 0   0 14 1 0   0 8 0 0",
"22 23 18     1 0 2 0   0 0 0 0  2 0 2 0    12 5 13 1  0 0 0 0",
"23 20 17     0 0 3 0   0 0 2 0  0 4 11 0   0 0 10 0   0 4 9 0",
"24 25 28     1 2 3 0   0 0 0 0  4 7 5 1    0 10 0 0   0 8 0 0",
"25 26 31     0 0 3 0   0 2 0 0  0 10 0 0   0 14 5 1   0 8 0 0",
"26 27 30     1 0 3 0   0 0 0 0  0 0 2 0    0 0 10 0   4 5 13 1",
"27 24 29     0 0 3 2   0 0 0 2  0 4 5 11   0 0 0 10   0 0 0 8",
"28 29 24     1 1 3 0   0 0 0 0  4 5 7 1    0 0 10 0   0 0 8 0",
"29 30 27     0 0 3 1   0 2 0 0  0 14 5 1   0 10 0 0   0 8 0 0",
"30 31 26     1 0 3 0   0 0 0 0  0 2 0 0    0 10 0 0   4 13 5 1",
"31 28 25     0 0 3 0   0 0 0 2  0 0 0 10   0 4 5 11   0 0 0 8",
"32 33 36     1 2 3 0   0 0 0 0  0 2 0 0    6 13 5 1   8 0 0 0",
"33 34 39     0 0 3 0   0 0 2 0  0 0 10 0   0 4 11 0   0 0 12 1",
"34 35 38     1 0 3 0   0 0 0 0  0 0 0 2    4 5 7 9    0 0 8 0",
"35 32 37     0 0 3 2   0 4 3 0  0 0 14 1   0 0 10 0   0 0 8 0",
"36 37 32     1 1 3 0   0 0 0 0  0 0 2 0    4 5 13 3   0 0 0 8",
"37 38 35     0 0 3 1   0 0 6 1  0 4 11 0   0 0 10 0   0 0 8 0",
"38 39 34     1 0 3 0   0 0 0 0  2 0 0 0    12 7 5 1   0 8 0 0",
"39 36 33     0 0 3 0   0 0 2 0  0 0 10 0   0 0 14 1   0 4 9 0",
"40 41 44     1 2 3 0   0 0 0 0  0 0 2 0    6 5 13 1   8 0 0 0",
"41 42 47     0 0 3 0   0 0 2 0  0 4 11 0   0 0 10 0   0 0 12 1",
"42 43 46     1 0 3 0   0 0 0 0  0 0 0 2    4 7 5 9    0 8 0 0",
"43 40 45     0 0 3 2   0 4 3 0  0 0 10 0   0 0 14 1   0 0 8 0",
"44 45 40     1 1 3 0   0 0 0 0  0 2 0 0    4 13 5 3   0 0 0 8",
"45 46 43     0 0 3 1   0 0 6 1  0 0 10 0   0 4 11 0   0 0 8 0",
"46 47 42     1 0 3 0   0 0 0 0  2 0 0 0    12 5 7 1   0 0 8 0",
"47 44 41     0 0 3 0   0 0 2 0  0 0 14 1   0 0 10 0   0 4 9 0",
"48 49 52     1 2 2 0   0 0 0 0  6 5 3 0    8 0 12 1   0 0 0 0",
"49 50 55     0 0 3 0   0 0 2 0  0 6 9 0    0 10 0 0   0 12 1 0",
"50 51 54     1 0 2 0   0 0 0 0  4 3 0 2    0 12 5 9   0 0 0 0",
"51 48 53     0 0 3 2   0 4 3 0  0 0 10 0   0 6 9 0    0 8 0 0",
"52 53 48     1 1 2 0   0 0 0 0  0 6 5 3    4 9 0 8    0 0 0 0",
"53 54 51     0 0 3 1   0 6 1 0  0 10 0 0   0 12 3 0   0 0 8 0",
"54 55 50     1 0 2 0   0 0 0 0  2 0 6 1    12 5 9 0   0 0 0 0",
"55 52 49     0 0 3 0   0 2 0 0  0 12 3 0   0 0 10 0   0 4 9 0",
"56 57 60     1 2 3 0   0 0 0 0  4 5 3 0    0 0 10 0   0 0 12 1",
"57 58 63     0 0 3 2   0 0 0 2  0 6 5 9    0 10 0 0   0 8 0 0",
"58 59 62     1 0 3 0   0 0 0 0  4 3 0 0    0 10 0 0   0 12 5 1",
"59 56 61     0 0 3 0   0 0 0 2  0 0 0 10   0 6 5 9    0 8 0 0",
"60 61 56     1 1 3 0   0 0 0 0  0 6 5 1    0 10 0 0   4 9 0 0",
"61 62 59     0 0 3 0   0 2 0 0  0 10 0 0   0 12 5 3   0 0 0 8",
"62 63 58     1 0 3 0   0 0 0 0  0 0 6 1    0 0 10 0   4 5 9 0",
"63 60 57     0 0 3 1   0 2 0 0  0 12 5 3   0 0 0 10   0 0 0 8",
"64 65 68     1 2 3 0   0 0 0 0  4 5 3 0    0 0 14 1   0 0 8 0",
"65 66 71     0 0 3 2   0 0 2 0  0 6 13 1   0 10 0 0   0 8 0 0",
"66 67 70     1 0 3 0   0 0 0 0  0 2 0 0    4 11 0 0   0 12 5 1",
"67 64 69     0 0 3 0   0 0 0 2  0 0 0 10   0 4 7 9    0 0 8 0",
"68 69 64     1 1 3 0   0 0 0 0  0 6 5 1    4 11 0 0   0 8 0 0",
"69 70 67     0 0 3 0   0 2 0 0  0 10 0 0   0 12 7 1   0 0 8 0",
"70 71 66     1 0 3 0   0 0 0 0  0 0 2 0    0 0 14 1   4 5 9 0",
"71 68 65     0 0 3 1   0 0 2 0  0 4 13 3   0 0 0 10   0 0 0 8",
"72 73 76     1 2 3 0   0 0 0 0  4 5 3 0    0 0 12 3   0 0 0 8",
"73 74 79     0 0 3 2   0 0 6 1  0 6 9 0    0 10 0 0   0 8 0 0",
"74 75 78     1 0 3 0   0 0 0 0  2 0 0 0    12 3 0 0   0 12 5 1",
"75 72 77     0 0 3 0   0 0 0 2  0 0 0 10   0 0 6 9    0 4 9 0",
"76 77 72     1 1 3 0   0 0 0 0  0 6 5 1    6 9 0 0    8 0 0 0",
"77 78 75     0 0 3 0   0 2 0 0  0 10 0 0   0 12 3 0   0 0 12 1",
"78 79 74     1 0 3 0   0 0 0 0  0 0 0 2    0 0 6 9    4 5 9 0",
"79 76 73     0 0 3 1   0 4 3 0  0 0 12 3   0 0 0 10   0 0 0 8",
"80 81 84     1 2 3 0   0 0 0 0  4 3 0 0    0 14 5 1   0 8 0 0",
"81 82 87     0 0 3 0   0 0 2 0  0 0 10 0   0 6 13 1   0 8 0 0",
"82 83 86     1 0 3 0   0 0 0 0  0 0 2 0    4 5 11 0   0 0 12 1",
"83 80 85     0 0 3 2   0 0 0 2  0 4 7 9    0 0 10 0   0 0 8 0",
"84 85 80     1 1 3 0   0 0 0 0  0 0 6 1    4 5 11 0   0 0 8 0",
"85 86 83     0 0 3 1   0 2 0 0  0 12 7 1   0 0 10 0   0 0 8 0",
"86 87 82     1 0 3 0   0 0 0 0  0 2 0 0    0 14 5 1   4 9 0 0",
"87 84 81     0 0 3 0   0 0 2 0  0 0 10 0   0 4 13 3   0 0 0 8",
"88 89 92     1 2 3 0   0 0 0 0  4 3 0 0    0 12 7 1   0 0 8 0",
"89 90 95     0 0 3 0   0 0 2 0  0 0 14 1   0 6 9 0    0 8 0 0",
"90 91 94     1 0 3 0   0 0 0 0  0 2 0 0    4 13 3 0   0 0 12 1",
"91 88 93     0 0 3 2   0 0 0 2  0 0 6 9    0 4 11 0   0 0 8 0",
"92 93 88     1 1 3 0   0 0 0 0  0 0 6 1    4 7 9 0    0 8 0 0",
"93 94 91     0 0 3 1   0 2 0 0  0 12 3 0   0 0 14 1   0 0 8 0",
"94 95 90     1 0 3 0   0 0 0 0  0 0 2 0    0 6 13 1   4 9 0 0",
"95 92 89     0 0 3 0   0 0 2 0  0 4 11 0   0 0 12 3   0 0 0 8",
"96 97 100    1 2 3 0   0 0 0 0  4 3 0 0    0 12 5 3   0 0 0 8",
"97 98 103    0 0 3 2   0 0 6 1  0 0 10 0   0 6 9 0    0 8 0 0",
"98 99 102    1 0 3 0   0 0 0 0  2 0 0 0    12 5 3 0   0 0 12 1",
"99 96 101    0 0 3 0   0 0 0 2  0 0 6 9    0 0 10 0   0 4 9 0",
"100 101 96   1 1 3 0   0 0 0 0  0 0 6 1    6 5 9 0    8 0 0 0",
"101 102 99   0 0 3 0   0 2 0 0  0 12 3 0   0 0 10 0   0 0 12 1",
"102 103 98   1 0 3 0   0 0 0 0  0 0 0 2    0 6 5 9    4 9 0 0",
"103 100 97   0 0 3 1   0 4 3 0  0 0 10 0   0 0 12 3   0 0 0 8",
"104 105 104  1 2 2 0   0 0 0 0  4 7 7 1    0 12 9 0   0 0 0 0",
"105 106 107  0 0 3 2   0 2 0 0  0 14 3 0   0 14 9 0   0 8 0 0",
"106 107 106  1 0 2 0   0 0 0 0  0 6 3 0    4 13 13 1  0 0 0 0",
"107 104 105  0 0 3 0   0 0 2 0  0 6 11 0   0 12 11 0  0 0 8 0",
"108 109 108  1 2 2 0   0 0 0 0  6 5 5 3    8 0 0 8    0 0 0 0",
"109 110 111  0 0 3 2   0 6 1 0  0 10 0 0   0 10 0 0   0 12 1 0",
"110 111 110  1 0 2 0   0 0 0 0  2 0 0 2    12 5 5 9   0 0 0 0",
"111 108 109  0 0 3 0   0 4 3 0  0 0 10 0   0 0 10 0   0 4 9 0",
"112 113 114  1 2 2 0   0 0 0 0  4 7 3 0    0 12 13 1  0 0 0 0",
"113 112 115  0 0 3 2   0 0 2 0  0 6 11 0   0 14 9 0   0 8 0 0",
"114 115 112  1 1 2 0   0 0 0 0  0 6 7 1    4 13 9 0   0 0 0 0",
"115 114 113  0 0 3 1   0 2 0 0  0 14 3 0   0 12 11 0  0 0 8 0",
"116 117 118  1 2 3 0   0 0 0 0  4 3 0 0    0 12 3 0   0 0 12 1",
"117 116 119  0 0 3 2   0 0 0 2  0 0 6 9    0 6 9 0    0 8 0 0",
"118 119 116  1 1 3 0   0 0 0 0  0 0 6 1    0 6 9 0    4 9 0 0",
"119 118 117  0 0 3 1   0 2 0 0  0 12 3 0   0 0 12 3   0 0 0 8",
"120 121 122  1 2 3 0   0 0 0 0  2 0 0 0    14 5 5 1   8 0 0 0",
"121 122 121  0 0 3 0   0 0 2 0  0 0 10 0   0 0 10 0   0 4 13 1",
"122 123 120  1 0 3 0   0 0 0 0  0 0 0 2    4 5 5 11   0 0 0 8",
"123 120 123  0 0 3 2   0 4 7 1  0 0 10 0   0 0 10 0   0 0 8 0",
"124 125 126  1 2 3 0   0 0 0 0  0 2 0 0    4 15 5 1   0 8 0 0",
"125 126 125  0 0 3 0   0 0 2 0  0 0 10 0   0 4 15 1   0 0 8 0",
"126 127 124  1 0 3 0   0 0 0 0  0 0 2 0    4 5 15 1   0 0 8 0",
"127 124 127  0 0 3 2   0 0 2 0  0 4 15 1   0 0 10 0   0 0 8 0",
"128 129 130  1 2 3 0   0 0 0 0  0 2 0 0    4 13 7 1   0 0 8 0",
"129 128 131  0 0 3 2   0 0 2 0  0 0 14 1   0 4 11 0   0 0 8 0",
"130 131 128  1 1 3 0   0 0 0 0  0 0 2 0    4 7 13 1   0 8 0 0",
"131 128 129  0 0 3 1   0 0 2 0  0 4 11 0   0 0 14 1   0 0 8 0",
"132 133 134  1 2 3 0   0 0 0 0  2 0 0 0    12 5 5 3   0 0 0 8",
"133 132 135  0 0 3 2   0 0 6 1  0 0 10 0   0 0 10 0   0 4 9 0",
"134 135 132  1 1 3 0   0 0 0 0  0 0 0 2    6 5 5 9    8 0 0 0",
"135 134 133  0 0 3 1   0 4 3 0  0 0 10 0   0 0 10 0   0 0 12 1",
"* 5",
"# 5 2",
"0 1 0  2 2 2 0   0 0 0 0 0  0 0 0 0 0   4 5 5 5 1   0 0 0 0 0   0 0 0 0 0",
"1 0 1  0 0 4 2   0 0 2 0 0  0 0 10 0 0  0 0 10 0 0  0 0 10 0 0  0 0 8 0 0",
"# 6 32",
"0 1 4     2 2 3 0   0 0 0 0 0  0 0 0 0 0   6 5 5 5 1   8 0 0 0 0   0 0 0 0 0",
"1 2 7     0 0 4 0   0 0 2 0 0  0 0 10 0 0  0 0 10 0 0  0 0 10 0 0  0 0 12 1 0",
"2 3 6     2 0 3 0   0 0 0 0 0  0 0 0 0 0   0 0 0 0 2   4 5 5 5 9   0 0 0 0 0",
"3 0 5     0 0 4 2   0 0 4 3 0  0 0 0 10 0  0 0 0 10 0  0 0 0 10 0  0 0 0 8 0",
"4 5 0     2 1 3 0   0 0 0 0 0  0 0 0 0 0   4 5 5 5 3   0 0 0 0 8   0 0 0 0 0",
"5 6 3     0 0 4 1   0 0 6 1 0  0 0 10 0 0  0 0 10 0 0  0 0 10 0 0  0 0 8 0 0",
"6 7 2     2 0 3 0   0 0 0 0 0  0 0 0 0 0   2 0 0 0 0   12 5 5 5 1  0 0 0 0 0",
"7 4 1     0 0 4 0   0 0 0 2 0  0 0 0 10 0  0 0 0 10 0  0 0 0 10 0  0 0 4 9 0",
"8 9 12    2 2 3 0   0 0 0 0 0  0 0 0 0 0   4 7 5 5 1   0 8 0 0 0   0 0 0 0 0",
"9 10 15   0 0 4 0   0 0 2 0 0  0 0 10 0 0  0 0 10 0 0  0 0 14 1 0  0 0 8 0 0",
"10 11 14  2 0 3 0   0 0 0 0 0  0 0 0 0 0   0 0 0 2 0   4 5 5 13 1  0 0 0 0 0",
"11 8 13   0 0 4 2   0 0 0 2 0  0 0 4 11 0  0 0 0 10 0  0 0 0 10 0  0 0 0 8 0",
"12 13 8   2 1 3 0   0 0 0 0 0  0 0 0 0 0   4 5 5 7 1   0 0 0 8 0   0 0 0 0 0",
"13 14 11  0 0 4 1   0 0 2 0 0  0 0 14 1 0  0 0 10 0 0  0 0 10 0 0  0 0 8 0 0",
"14 15 10  2 0 3 0   0 0 0 0 0  0 0 0 0 0   0 2 0 0 0   4 13 5 5 1  0 0 0 0 0",
"15 12 9   0 0 4 0   0 0 0 2 0  0 0 0 10 0  0 0 0 10 0  0 0 4 11 0  0 0 0 8 0",
"16 17 20  2 2 3 0   0 0 0 0 0  0 0 0 0 0   4 5 5 3 0   0 0 0 12 1  0 0 0 0 0",
"17 18 23  0 0 4 2   0 0 0 2 0  0 0 6 9 0   0 0 10 0 0  0 0 10 0 0  0 0 8 0 0",
"18 19 22  2 0 3 0   0 0 0 0 0  0 0 0 0 0   4 3 0 0 0   0 12 5 5 1  0 0 0 0 0",
"19 16 21  0 0 4 0   0 0 0 2 0  0 0 0 10 0  0 0 0 10 0  0 0 6 9 0   0 0 8 0 0",
"20 21 16  2 1 3 0   0 0 0 0 0  0 0 0 0 0   0 6 5 5 1   4 9 0 0 0   0 0 0 0 0",
"21 22 19  0 0 4 0   0 0 2 0 0  0 0 10 0 0  0 0 10 0 0  0 0 12 3 0  0 0 0 8 0",
"22 23 18  2 0 3 0   0 0 0 0 0  0 0 0 0 0   0 0 0 6 1   4 5 5 9 0   0 0 0 0 0",
"23 20 17  0 0 4 1   0 0 2 0 0  0 0 12 3 0  0 0 0 10 0  0 0 0 10 0  0 0 0 8 0",
"24 25 24  2 2 3 0   0 0 0 0 0  0 0 0 0 0   4 5 7 5 1   0 0 8 0 0   0 0 0 0 0",
"25 26 27  0 0 4 2   0 0 2 0 0  0 0 10 0 0  0 0 14 1 0  0 0 10 0 0  0 0 8 0 0",
"26 27 26  2 0 3 0   0 0 0 0 0  0 0 0 0 0   0 0 2 0 0   4 5 13 5 1  0 0 0 0 0",
"27 24 25  0 0 4 0   0 0 0 2 0  0 0 0 10 0  0 0 4 11 0  0 0 0 10 0  0 0 0 8 0",
"28 29 30  2 2 3 0   0 0 0 0 0  0 0 0 0 0   4 5 3 0 0   0 0 12 5 1  0 0 0 0 0",
"29 28 31  0 0 4 2   0 0 0 2 0  0 0 0 10 0  0 0 6 9 0   0 0 10 0 0  0 0 8 0 0",
"30 31 28  2 1 3 0   0 0 0 0 0  0 0 0 0 0   0 0 6 5 1   4 5 9 0 0   0 0 0 0 0",
"31 30 29  0 0 4 1   0 0 2 0 0  0 0 10 0 0  0 0 12 3 0  0 0 0 10 0  0 0 0 8 0",
"* 6",
"# 6 2",
"0 1 0  3 2 3 0   0 0 0 0 0 0   0 0 0 0 0 0   0 0 0 0 0 0",
"                 4 5 5 5 5 1   0 0 0 0 0 0   0 0 0 0 0 0",
"1 0 1  0 0 5 2   0 0 0 2 0 0   0 0 0 10 0 0  0 0 0 10 0 0",
"                 0 0 0 10 0 0  0 0 0 10 0 0  0 0 0 8 0 0"};
}

class Spec {
    volatile int units;
    volatile boolean diagonal, mixed, bonus;

    public Spec(int units,
                    boolean diagonal, boolean mixed, boolean bonus) {
            this.units = units;
            this.diagonal = diagonal;
            this.mixed = mixed;
            this.bonus = bonus;
    }

    public Spec(int units) {
            this.units = units;
            this.diagonal = this.mixed = this.bonus = false;
    }

    void copyValues(Spec nextSpec) {
            units = nextSpec.units;
            diagonal = nextSpec.diagonal;
            mixed = nextSpec.mixed;
            bonus = nextSpec.bonus;
    }

    public void setDiagonal(boolean diagonal) {
            this.diagonal = diagonal;
    }

    public void setMixed(boolean mixed) {
            this.mixed = mixed;
    }

    public void setBonus(boolean bonus) {
        this.bonus = bonus;
    }

    public void setUnits(int units) {
        this.units = units;
    }
}

